// Welcome to
// __________         __    __  .__                               __
// \______   \_____ _/  |__/  |_|  |   ____   ______ ____ _____  |  | __ ____
//  |    |  _/\__  \\   __\   __\  | _/ __ \ /  ___//    \\__  \ |  |/ // __ \
//  |    |   \ / __ \|  |  |  | |  |_\  ___/ \___ \|   |  \/ __ \|    <\  ___/
//  |________/(______/__|  |__| |____/\_____>______>___|__(______/__|__\\_____>
//
// This file can be a nice home for your Battlesnake logic and helper functions.
//
// To get you started we've included code to prevent your Battlesnake from moving backwards.
// For more info see docs.battlesnake.com

use core::f64;
use std::{
    cell::Cell,
    collections::{BTreeMap, HashMap},
    sync::{Arc, LazyLock, Mutex},
    time::{Duration, Instant},
};

use log::{error, info};
use ordered_float::OrderedFloat;
use serde_json::{json, Value};

use crate::{
    simulation::{self, SnakeToken},
    Action, Battlesnake, Board, Direction, Game, MAX_HEALTH,
};

impl Battlesnake {
    fn possible_actions_without_heads<'a>(
        &'a self,
        game: &'a Game,
        board: &'a Board,
    ) -> impl Iterator<Item = Direction> + 'a {
        enum_iterator::all::<Direction>()
            .filter(|direction| {
                // filter out directions that would go outside the field
                let target = self.head.move_to(*direction);
                (0..board.width).contains(&target.x) && (0..board.height).contains(&target.y)
            })
            .filter(|direction| {
                let target = self.head.move_to(*direction);
                // don't collide with any snake
                !board
                    .snakes
                    .iter()
                    .filter(|snake| {
                        // filter out snakes that are in our squad if body collisions are allowed
                        !(game.ruleset.settings.squad.allow_body_collisions
                            && self.squad == snake.squad
                            && self.id != snake.id)
                    })
                    .flat_map(|snake| {
                        // get all coordinates of the snake body. The tail of the body can be ignored
                        // if the snake hasn't just eaten, as it will move out of the way
                        let has_eaten = snake.health == MAX_HEALTH;
                        snake.body[..snake.body.len() - usize::from(!has_eaten)].iter()
                    })
                    .any(|&coord| coord == target)
            })
    }

    #[must_use]
    pub fn possible_actions(&self, game: &Game, board: &Board) -> Vec<Direction> {
        self.possible_actions_without_heads(game, board)
            .filter(|direction| {
                // don't go into spots where a bigger snake must go
                let target = self.head.move_to(*direction);
                !board
                    .snakes
                    .iter()
                    .filter(|snake| {
                        // The other snake must be bigger than we are
                        snake.length > self.length
                    })
                    .filter_map(|snake| {
                        // get all snakes movement options
                        let actions = snake
                            .possible_actions_without_heads(game, board)
                            .collect::<Vec<_>>();
                        // only snakes that have a single option
                        match actions[..] {
                            [direction] => Some(snake.head.move_to(direction)),
                            _ => None,
                        }
                    })
                    .any(|coord| coord == target)
            })
            .collect()
    }
}

// info is called when you create your Battlesnake on play.battlesnake.com
// and controls your Battlesnake's appearance
// TIP: If you open your Battlesnake URL in a browser you should see this data
pub fn info() -> Value {
    info!("INFO");

    json!({
        "apiversion": "1",
        "author": "der-informatiker",
        "color": "#00FFEE",
        "head": "smart-caterpillar",
        "tail": "mouse",
    })
}

#[derive(Debug, PartialEq, Eq, Clone)]
struct GameInfo {
    calculation_time: Cell<Duration>,
    token_mapping: Arc<BTreeMap<String, SnakeToken>>,
    my_token: SnakeToken,
}

static GAME_INFOS: LazyLock<Mutex<HashMap<(String, String), GameInfo>>> =
    LazyLock::new(|| Mutex::new(HashMap::new()));

// start is called when your Battlesnake begins a game
pub fn start(game: &Game, _turn: i32, board: &Board, you: &Battlesnake) {
    info!("GAME START");
    let token_mapping = Arc::new(SnakeToken::from_board(board));
    let my_token = token_mapping[&you.id];
    let Ok(mut game_infos) = GAME_INFOS.lock() else {
        error!("unable to lock game infos");
        return;
    };
    game_infos.insert(
        (game.id.clone(), you.id.clone()),
        GameInfo {
            calculation_time: Cell::new(Duration::from_millis(50)),
            token_mapping,
            my_token,
        },
    );
}

// end is called when your Battlesnake finishes a game
pub fn end(game: &Game, _turn: i32, _board: &Board, you: &Battlesnake) {
    info!("GAME OVER");
    let Ok(mut game_infos) = GAME_INFOS.lock() else {
        error!("unable to lock game infos");
        return;
    };
    game_infos.remove(&(game.id.clone(), you.id.clone()));
}

// move is called on every turn and returns your next move
// Valid moves are "up", "down", "left", or "right"
// See https://docs.battlesnake.com/api/example-move for available data
pub fn get_move(game: &Game, turn: i32, board: &Board, you: &Battlesnake) -> Option<Action> {
    let start = Instant::now();
    let game_info = GAME_INFOS
        .lock()
        .ok()
        .and_then(|guard| guard.get(&(game.id.clone(), you.id.clone())).cloned())
        .unwrap_or_else(|| {
            let token_mapping = Arc::new(SnakeToken::from_board(board));
            let my_token = token_mapping[&you.id];
            GameInfo {
                calculation_time: Cell::new(Duration::from_millis(50)),
                token_mapping,
                my_token,
            }
        });

    // do some latency compensation
    game_info.calculation_time.set(
        game_info.calculation_time.get()
            + Duration::from_millis(
                u64::from(
                    game.timeout * 3 / 4 - you.latency.parse().unwrap_or(game.timeout * 3 / 4),
                ) / 3,
            ),
    );

    let board = simulation::Board::from_game_board(
        board,
        &game_info.token_mapping,
        turn,
        game.ruleset.settings.food_spawn_chance,
        game.ruleset.settings.minimum_food,
    );

    let mut tree = Node::default();

    while start.elapsed() < game_info.calculation_time.get() {
        let mut board = board.clone();
        if game.ruleset.name == "solo" {
            tree.monte_carlo_step_solo(&mut board);
        } else {
            tree.monte_carlo_step(&mut board);
        }
    }

    let actions = tree.child_statistics.entry(game_info.my_token).or_default();

    info!("actions {}: {actions:?}", you.name);

    #[allow(clippy::cast_precision_loss)]
    let chosen = actions
        .iter()
        .max_by_key(|(_, stat)| OrderedFloat(stat.won as f64 / stat.played as f64))
        .map(|(direction, _)| *direction)?;

    info!(
        "DIRECTION {turn}: {chosen:?} after {}ms ({})",
        start.elapsed().as_millis(),
        you.name,
    );
    Some(Action {
        r#move: chosen,
        shout: None,
    })
}

#[derive(Debug, PartialEq, Eq, Clone, Default)]
struct Statistics {
    /// Number of times this node was simulated
    played: usize,
    /// Number of times this node was simulated and the agent has won.
    won: BTreeMap<SnakeToken, usize>,
}

#[derive(Debug, PartialEq, Eq, Clone, Default)]
struct ActionStatistic {
    played: usize,
    won: usize,
}

#[derive(Debug, PartialEq, Eq, Clone, Default)]
struct Node {
    statistic: Statistics,
    child_statistics: BTreeMap<SnakeToken, BTreeMap<Direction, ActionStatistic>>,
    childs: BTreeMap<BTreeMap<SnakeToken, Direction>, Node>,
}

impl Node {
    /// Performs one monte carlo simulation step
    ///
    /// Returns the snake that has won the simulation
    fn monte_carlo_step(&mut self, board: &mut simulation::Board) -> Option<SnakeToken> {
        let winner = if self.statistic.played == 0 {
            // We didn't simulate a game for this node yet. Do that
            board.simulate_until(|board| board.alive_snakes() <= 1);
            board.snakes().next()
        } else {
            // select a node to simulate
            let possible_actions = board.possible_actions();

            let actions = possible_actions
                .iter()
                .filter_map(|(token, actions)| {
                    let statistics = self.child_statistics.entry(*token).or_default();
                    let selected = actions.iter().copied().max_by_key(|direction| {
                        let statistics = statistics.entry(*direction).or_default();
                        if statistics.played == 0 {
                            return OrderedFloat(f64::INFINITY);
                        }
                        #[allow(clippy::cast_precision_loss)]
                        let exploitation = statistics.won as f64 / statistics.played as f64;
                        #[allow(clippy::cast_precision_loss)]
                        let exploration = f64::consts::SQRT_2
                            * f64::sqrt(
                                f64::ln(self.statistic.played as f64) / statistics.played as f64,
                            );
                        OrderedFloat(exploitation + exploration)
                    })?;
                    Some((*token, selected))
                })
                .collect();

            board.simulate_actions(&actions);
            let winner = self
                .childs
                .entry(actions.clone())
                .or_default()
                .monte_carlo_step(board);

            // update child statistics
            for (token, action) in &actions {
                let entry = self
                    .child_statistics
                    .entry(*token)
                    .or_default()
                    .entry(*action)
                    .or_default();
                entry.played += 1;
                if Some(*token) == winner {
                    entry.won += 1;
                }
            }

            winner
        };
        self.statistic.played += 1;
        if let Some(token) = winner {
            self.statistic
                .won
                .entry(token)
                .and_modify(|won| *won += 1)
                .or_insert(1);
        }
        winner
    }

    /// Performs one monte carlo simulation step for a solo game
    ///
    /// Returns the lengths before death
    fn monte_carlo_step_solo(
        &mut self,
        board: &mut simulation::Board,
    ) -> BTreeMap<SnakeToken, usize> {
        let lengths = if self.statistic.played == 0 {
            // We didn't simulate a game for this node yet. Do that
            let mut lengths: BTreeMap<_, _> = board
                .snakes()
                .filter_map(|snake| Some((snake, board.snake_length(snake)?)))
                .collect();
            board.simulate_until(|board| {
                for snake in board.snakes() {
                    if let Some(length) = board.snake_length(snake) {
                        lengths.insert(snake, length);
                    }
                }
                board.alive_snakes() == 0
            });
            lengths
        } else {
            // select a node to simulate
            let possible_actions = board.possible_actions();

            let actions = possible_actions
                .iter()
                .filter_map(|(token, actions)| {
                    let statistics = self.child_statistics.entry(*token).or_default();
                    let selected = actions.iter().copied().max_by_key(|direction| {
                        let statistics = statistics.entry(*direction).or_default();
                        if statistics.played == 0 {
                            return OrderedFloat(f64::INFINITY);
                        }
                        #[allow(clippy::cast_precision_loss)]
                        let exploitation = statistics.won as f64
                            / board.spaces() as f64
                            / statistics.played as f64;
                        #[allow(clippy::cast_precision_loss)]
                        let exploration = f64::consts::SQRT_2
                            * f64::sqrt(
                                f64::ln(self.statistic.played as f64) / statistics.played as f64,
                            );
                        OrderedFloat(exploitation + exploration)
                    })?;
                    Some((*token, selected))
                })
                .collect();

            board.simulate_actions(&actions);
            let lengths = self
                .childs
                .entry(actions.clone())
                .or_default()
                .monte_carlo_step_solo(board);

            // update child statistics
            for (token, action) in &actions {
                let entry = self
                    .child_statistics
                    .entry(*token)
                    .or_default()
                    .entry(*action)
                    .or_default();
                entry.played += 1;
                if let Some(length) = lengths.get(&token) {
                    entry.won += length;
                }
            }

            lengths
        };
        self.statistic.played += 1;
        for (token, length) in &lengths {
            self.statistic
                .won
                .entry(*token)
                .and_modify(|won| *won += length)
                .or_insert(*length);
        }
        lengths
    }
}
